/* db_compiler.c - a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <string.h>
#include <ctype.h>
#include "db_compiler.h"
#include "db_vmdebug.h"
#include "db_vm.h"
#include "db_buffer.h"

/* local function prototypes */
static void PlaceStrings(ParseContext *c);

/* InitCompiler - initialize the compiler */
ParseContext *InitCompiler(System *sys)
{
    ParseContext *c;
    if (!(c = (ParseContext *)AllocateFreeSpace(sys, sizeof(ParseContext))))
        return NULL;
    memset(c, 0, sizeof(ParseContext));
    c->sys = sys;
    return c;
}

/* Compile - compile a program */
ImageHdr *Compile(ParseContext *c, int maxObjects)
{
	/* setup an error target */
    if (setjmp(c->errorTarget) != 0)
        return NULL;
        
    /* start the image */
    StartImage(c);

    /* allocate space for the object table */
    if (!(c->objects = (VMUVALUE *)AllocateFreeSpace(c->sys, maxObjects * sizeof(VMVALUE))))
        ParseError(c, "insufficient space for object table\n");
    c->objectCount = 0;
    c->maxObjects = maxObjects;
    c->objectDataSize = 0;

    /* use the rest of the free space for the compiler heap */
    c->nextGlobal = AllocateAllFreeSpace(c->sys, &c->heapSize);
    c->heapTop = c->nextLocal = c->nextGlobal + c->heapSize;
    c->maxHeapUsed = 0;

    /* initialize the global variable count */
    c->variableCount = 0;

    /* initialize block nesting table */
    c->btop = (Block *)((char *)c->blockBuf + sizeof(c->blockBuf));
    c->bptr = c->blockBuf - 1;

    /* initialize the code staging buffer */
    c->ctop = c->codeBuf + sizeof(c->codeBuf);
    c->cptr = c->codeBuf;

    /* initialize the string and label tables */
    c->strings = NULL;
    c->labels = NULL;

    /* initialize the global symbol table */
    InitSymbolTable(&c->globals);

    /* add the intrinsic functions */
    AddIntrinsic(c, "ABS",          "i",    "i",    FN_ABS);
    AddIntrinsic(c, "RND",          "i",    "i",    FN_RND);
    AddIntrinsic(c, "LEFT$",        "si",   "s",    FN_LEFT);
    AddIntrinsic(c, "RIGHT$",       "si",   "s",    FN_RIGHT);
    AddIntrinsic(c, "MID$",         "sii",  "s",    FN_MID);
    AddIntrinsic(c, "CHR$",         "i",    "s",    FN_CHR);
    AddIntrinsic(c, "STR$",         "i",    "s",    FN_STR);
    AddIntrinsic(c, "ASC",          "s",    "i",    FN_ASC);
    AddIntrinsic(c, "LEN",          "s",    "i",    FN_LEN);
    
    /* add the registers */
    AddRegister(c, "CNT",           RG_CNT);
    AddRegister(c, "DIRA",          RG_DIRA);
    AddRegister(c, "OUTA",          RG_OUTA);
    AddRegister(c, "INA",           RG_INA);
    AddRegister(c, "CLKFREQ",       RG_CLKFREQ);

    /* initialize scanner */
    c->inComment = VMFALSE;
    
    /* start in the main code */
    c->codeType = NULL;

    /* set the compiler pass number */
    c->pass = 1;
    
    /* do three passes over the source program */
    for (c->pass = 1; c->pass <= 3; ++c->pass) {
        
        /* rewind to the start of the source program */
        (*c->rewind)(c->getLineCookie);
    
        /* get the next line */
        while (GetLine(c)) {
            Token tkn;
            if ((tkn = GetToken(c)) != T_EOL)
                ParseStatement(c, tkn);
        }
    }

    /* end the main code with a halt */
    putcbyte(c, OP_HALT);
    
    /* write the main code */
    StartCode(c, "main", NULL, NewObject(c));
    c->mainCode = c->code;
    StoreCode(c);

#if 0
    DumpSymbols(&c->globals, "symbols");
#endif

    /* build an image in memory */
    return BuildImage(c);
}

/* StartCode - start a function or method under construction */
void StartCode(ParseContext *c, char *name, Type *type, VMVALUE code)
{
    /* initialize the code object under construction */
    c->codeName = name;
    c->codeType = NULL;
    InitSymbolTable(&c->locals);
    c->localOffset = 0;
    c->codeType = type;
    c->code = code;
    
    /* write the code prolog */
    if (type) {
        putcbyte(c, OP_RESERVE);
        putcbyte(c, 0);
    }
}

/* StoreCode - store the function or method under construction */
void StoreCode(ParseContext *c)
{
    int codeSize;

    /* check for unterminated blocks */
	switch (CurrentBlockType(c)) {
    case BLOCK_IF:
    case BLOCK_ELSE:
		ParseError(c, "expecting END IF");
    case BLOCK_FOR:
		ParseError(c, "expecting NEXT");
	case BLOCK_DO:
		ParseError(c, "expecting LOOP");
	case BLOCK_NONE:
		break;
	}

    /* fixup the RESERVE instruction at the start of the code */
    if (c->codeType) {
        c->codeBuf[1] = c->localOffset;
        putcbyte(c, OP_RETURN);
    }

    /* make sure all referenced labels were defined */
    CheckLabels(c);

    /* place string literals defined in this function */
    PlaceStrings(c);
    
    /* determine the code size */
    codeSize = c->cptr - c->codeBuf;

#if 0
    VM_printf("%s:\n", c->codeName);
    DecodeFunction(c->code, c->codeBuf, codeSize);
    if (c->codeType)
        DumpSymbols(&c->codeType->u.functionInfo.arguments, "arguments");
    DumpSymbols(&c->locals, "locals");
    VM_printf("\n");
#endif

    /* store the vector object */
    StoreBVectorData(c, c->code, PROTO_CODE, c->codeBuf, codeSize);

    /* empty the local heap */
    c->nextLocal = c->heapTop;
    
    /* empty the symbol and label tables */
    InitSymbolTable(&c->locals);
    c->labels = NULL;

    /* reset to compile the next code */
    c->cptr = c->codeBuf;
    c->codeType = NULL;
}

/* PlaceStrings - place any string literals defined in the current function */
static void PlaceStrings(ParseContext *c)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next)
        if (str->fixups)
            StoreString(c, str);
}

/* AddString - add a string to the string table */
String *AddString(ParseContext *c, char *value)
{
    String *str;
    
    /* check to see if the string is already in the table */
    for (str = c->strings; str != NULL; str = str->next)
        if (strcmp(value, (char *)str->value) == 0)
            return str;

    /* allocate the string structure */
    str = (String *)GlobalAlloc(c, sizeof(String) + strlen(value) - 1);
    memset(str, 0, sizeof(String));
    strcpy((char *)str->value, value);
    str->next = c->strings;
    c->strings = str;

    /* return the string table entry */
    return str;
}

/* AddStringRef - add a reference to a string in the string table */
VMVALUE AddStringRef(String *str, int offset)
{
    int link;

    /* handle strings that have already been placed */
    if (str->placed)
        return str->object;

    /* add a new entry to the fixup list */
    link = str->fixups;
    str->fixups = offset;
    return link;
}

/* StoreString - store a string */
VMVALUE StoreString(ParseContext *c, String *str)
{
    /* handle strings that have not already been placed */
    if (!str->placed) {
        str->object = StoreBVector(c, str->value, strlen((char *)str->value));
        str->placed = VMTRUE;
        fixup(c, str->fixups, str->object);
        str->fixups = 0;
    }
    
    /* return the stirng object */
    return str->object;
}

/* AddIntrinsic - add an intrinsic function to the global symbol table */
void AddIntrinsic(ParseContext *c, char *name, char *argTypes, char *retType, int index)
{
    Type *type, *argType;
    VMUVALUE off;

    /* create a type for the function */
    type = NewGlobalType(c, TYPE_FUNCTION);
    InitSymbolTable(&type->u.functionInfo.arguments);
    
    /* set the codeType for AddArgument */
    c->codeType = type;

    /* set the return type */
    switch (*retType) {
    case 'i':
        type->u.functionInfo.returnType = &integerType;
        break;
    case 's':
        type->u.functionInfo.returnType = &stringType;
        break;
    default:
        Fatal(c, "unknown return type: %s", retType);
        break;
    }

    /* add the argument types */
    for (off = 0; *argTypes; ++off) {
        switch (*argTypes++) {
        case 'i':
            argType = &integerType;
            break;
        case 's':
            argType = &stringType;
            break;
        default:
            Fatal(c, "unknown argument type: %c", argTypes[-1]);
            argType = NULL; // never reached
            break;
        }
        AddGlobalSymbol(c, &type->u.functionInfo.arguments, "", SC_VARIABLE, argType, off);
    }
    
    /* add a global symbol for the intrinsic function */
    AddGlobal(c, name, SC_CONSTANT, type, INTRINSIC_FLAG, INTRINSIC_FLAG | index);
}

/* AddRegister - add a register to the global symbol table */
void AddRegister(ParseContext *c, char *name, int index)
{
    AddGlobal(c, name, SC_REGISTER, &integerType, 0, index);
}

/* GlobalAlloc - allocate memory from the global heap */
void *GlobalAlloc(ParseContext *c, size_t size)
{
    void *p;
    size = (size + ALIGN_MASK) & ~ALIGN_MASK;
    if (c->nextGlobal + size > c->nextLocal)
        Fatal(c, "insufficient memory");
    p = c->nextGlobal;
    c->nextGlobal += size;
    if (c->heapSize - (c->nextLocal - c->nextGlobal) > c->maxHeapUsed)
        c->maxHeapUsed = c->heapSize - (c->nextLocal - c->nextGlobal);
    return p;
}

/* LocalAlloc - allocate memory from the local heap */
void *LocalAlloc(ParseContext *c, size_t size)
{
    size = (size + ALIGN_MASK) & ~ALIGN_MASK;
    if (c->nextLocal - size < c->nextGlobal)
        Fatal(c, "insufficient memory");
    c->nextLocal -= size;
    if (c->heapSize - (c->nextLocal - c->nextGlobal) > c->maxHeapUsed)
        c->maxHeapUsed = c->heapSize - (c->nextLocal - c->nextGlobal);
    return c->nextLocal;
}

/* VM_printf - formatted print */
void VM_printf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_vprintf(fmt, ap);
    va_end(ap);
}

/* Fatal - report a fatal error and exit */
void Fatal(ParseContext *c, char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf("error: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    longjmp(c->errorTarget, 1);
}
