/* db_image.c - compiled image functions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include "db_compiler.h"
#include "db_vmdebug.h"
#include "db_buffer.h"
#include "db_vm.h"

/* global initialization entry */
typedef struct {
    VMUVALUE offset;
    VMVALUE value;
} GlobalInitEntry;

/* StartImage - start an image */
void StartImage(ParseContext *c)
{
    ScratchRewind();
}

/* BuildImage - build an image from the symbol table and objects already written to the scratch file */
ImageHdr *BuildImage(ParseContext *c)
{
    System *sys = c->sys;
    VMVALUE mainCode;
    VMUVALUE variableCount, objectCount, objectDataSize;
    int initEntryCount = 0, n;
    GlobalInitEntry entry;
    ImageHdr *image;
    Symbol *sym;
    
    {
        int objectTableSize = c->objectCount  * sizeof(VMUVALUE);
        int dataSize = objectTableSize + c->objectDataSize + c->variableCount * sizeof(VMVALUE);
        VM_printf("H:%d", c->maxHeapUsed);
        VM_printf(" O:%d", c->objectCount);
        VM_printf(" D:%d", c->objectDataSize);
        VM_printf(" V:%d", c->variableCount);
        VM_printf(" T:%d\n", dataSize);
    }
    
    /* write the initial values of the global variables to the scratch file */
    for (sym = c->globals.head; sym != NULL; sym = sym->next)
        if (sym->storageClass == SC_VARIABLE) {
            entry.offset = sym->offset;
            entry.value = sym->initialValue.ivalue;
            if (!ScratchWrite((uint8_t *)&entry, sizeof(entry)))
                ParseError(c, "insufficient object scratch space");
            ++initEntryCount;
        }
        
    /* write the object table to the scratch file */
    if (!ScratchWrite((uint8_t *)c->objects, c->objectCount * sizeof(VMUVALUE)))
        Fatal(c, "insufficient object scratch space");
        
    /* get the information we need out of the parse context so we can overwrite it */
    mainCode = c->mainCode; 
    variableCount = c->variableCount;
    objectCount = c->objectCount;
    objectDataSize = c->objectDataSize;

    /* deallocate everything including the ParseContext! */
    ResetFreeSpace(sys);
        
    /* allocate space for the image header */
    if (!(image = (ImageHdr *)AllocateFreeSpace(sys, sizeof(ImageHdr))))
        Fatal(c, "insufficient space for the image header");
        
    /* initialize the image */
    image->mainCode = mainCode;
    image->variableCount = variableCount;
    
    /* initialize the heap */
    if (!InitHeap(sys, image, objectDataSize + 256, objectCount + 8))
        Fatal(c, "insufficient space for the heap");
        
    /* reserve space in the heap for the preallocated objects */
    ReserveHeapSpace(image, objectDataSize, objectCount);
    
    /* allocate the global variable table */
    if (!(image->variables = (VMVALUE *)AllocateFreeSpace(sys, image->variableCount * sizeof(VMVALUE))))
        Fatal(c, "insufficient space for variable table");

    /* rewind the scratch file */
    ScratchRewind();
    
    /* read the object data */
    if (!ScratchRead(image->data, objectDataSize))
        Fatal(c, "error reading object scratch space");

    /* read the initial values of the global variables */
    while (--initEntryCount >= 0) {
        if (!ScratchRead((uint8_t *)&entry, sizeof(entry)))
            Fatal(c, "error reading object scratch space");
        image->variables[entry.offset] = entry.value;
    }
    
    /* read the object data offsets */
    for (n = 1; n <= objectCount; ++n) {
        ObjectHandle *handle;
        VMUVALUE offset;
        if (!ScratchRead((uint8_t *)&offset, sizeof(offset)))
            Fatal(c, "error reading object scratch space");
        handle = GetObjHandle(image, n);
        handle->hdr = (ObjectHdr *)(image->data + offset);
        handle->hdr->handle = handle;
    }
    
#if 0
    DumpHeap(image);
#endif
    
    /* return the image */
    return image;
}

/* StoreVector - store a vector object */
VMVALUE StoreVector(ParseContext *c, const VMVALUE *buf, int size)
{
    int dataSizeInBytes = size * sizeof(VMVALUE);
    ObjectHdr hdr;
    VMVALUE object;

    /* allocate and initialize the intrinsic function object */
    object = NewObject(c);

    /* store the object offset */
    c->objects[object - 1] = c->objectDataSize;

    /* write the object header and data */
    hdr.handle = NULL;
    hdr.prototype = PROTO_VECTOR;
    hdr.size = size;
    if (!ScratchWrite((uint8_t *)&hdr, sizeof(hdr))
    ||  !ScratchWrite((uint8_t *)buf, dataSizeInBytes))
        ParseError(c, "insufficient object scratch space");

    /* update the object data size */
    c->objectDataSize += sizeof(hdr) + dataSizeInBytes;

    /* return the object */
    return object;
}

/* StoreBVector - store a byte vector object */
VMVALUE StoreBVector(ParseContext *c, const uint8_t *buf, int size)
{
    VMVALUE object;

    /* allocate and initialize the intrinsic function object */
    object = NewObject(c);

    /* store the vector object */
    StoreBVectorData(c, object, PROTO_BVECTOR, buf, size);

    /* return the object */
    return object;
}

/* StoreBVectorData - store data for a byte vector object */
void StoreBVectorData(ParseContext *c, VMVALUE object, VMVALUE proto, const uint8_t *buf, int size)
{
    int allocatedSize = (size + ALIGN_MASK) & ~ALIGN_MASK;
    ObjectHdr hdr;

    /* store the object offset */
    c->objects[object - 1] = c->objectDataSize;

    /* write the object header and data */
    hdr.handle = NULL;
    hdr.prototype = proto;
    hdr.size = size;
    if (!ScratchWrite((uint8_t *)&hdr, sizeof(hdr))
    ||  !ScratchWrite((uint8_t *)buf, allocatedSize))
        ParseError(c, "insufficient object scratch space");

    /* update the object data size */
    c->objectDataSize += sizeof(hdr) + allocatedSize;
}

/* NewObject - create a new object */
VMVALUE NewObject(ParseContext *c)
{
    VMVALUE object;
    if (c->objectCount >= c->maxObjects)
        ParseError(c, "too many objects");
    object = ++c->objectCount;
    c->objects[object - 1] = 0;
    return object;
}
