/* db_symbols.c - symbol table routines
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "db_compiler.h"

/* InitSymbolTable - initialize a symbol table */
void InitSymbolTable(SymbolTable *table)
{
    table->head = NULL;
    table->pTail = &table->head;
    table->count = 0;
}

/* AddGlobal - add a global symbol to the symbol table */
Symbol *AddGlobal(ParseContext *c, const char *name, StorageClass storageClass, Type *type, VMUVALUE offset, VMVALUE initialValue)
{
    Symbol *sym = AddGlobalSymbol(c, &c->globals, name, storageClass, type, offset);
    sym->initialValue.ivalue = initialValue;
    return sym;
}

/* AddGlobalConstantInteger - add a global constant integer symbol to the symbol table */
Symbol *AddGlobalConstantInteger(ParseContext *c, const char *name, VMVALUE value)
{
    Symbol *sym = AddGlobalSymbol(c, &c->globals, name, SC_CONSTANT, &integerType, 0);
    sym->initialValue.ivalue = value;
    return sym;
}

/* AddGlobalConstantString - add a global constant string symbol to the symbol table */
Symbol *AddGlobalConstantString(ParseContext *c, const char *name, String *string)
{
    Symbol *sym = AddGlobalSymbol(c, &c->globals, name, SC_CONSTANT, &stringType, 0);
    sym->initialValue.string = string;
    return sym;
}

/* AddGlobalSymbol - add a symbol to a global symbol table */
Symbol *AddGlobalSymbol(ParseContext *c, SymbolTable *table, const char *name, StorageClass storageClass, Type *type, VMUVALUE offset)
{
    size_t size = sizeof(Symbol) + strlen(name);
    Symbol *sym;
    
    /* allocate the symbol structure */
    sym = (Symbol *)GlobalAlloc(c, size);
    strcpy(sym->name, name);
    sym->storageClass = storageClass;
    sym->type = type;
    sym->offset = offset;
    sym->initialValue.ivalue = 0;
    sym->next = NULL;

    /* add it to the symbol table */
    *table->pTail = sym;
    table->pTail = &sym->next;
    ++table->count;
    
    /* return the symbol */
    return sym;
}

/* AddLocal - add a local symbol to the symbol table */
Symbol *AddLocal(ParseContext *c, const char *name, Type *type, VMUVALUE offset)
{
    SymbolTable *table = &c->locals;
    size_t size = sizeof(Symbol) + strlen(name);
    Symbol *sym;
    
    /* allocate the symbol structure */
    sym = (Symbol *)LocalAlloc(c, size);
    strcpy(sym->name, name);
    sym->storageClass = SC_VARIABLE;
    sym->type = type;
    sym->offset = offset;
    sym->initialValue.ivalue = 0;
    sym->next = NULL;

    /* add it to the symbol table */
    *table->pTail = sym;
    table->pTail = &sym->next;
    ++table->count;
    
    /* return the symbol */
    return sym;
}

/* FindSymbol - find a symbol in a symbol table */
Symbol *FindSymbol(SymbolTable *table, const char *name)
{
    Symbol *sym = table->head;
    while (sym) {
        if (strcasecmp(name, sym->name) == 0)
            return sym;
        sym = sym->next;
    }
    return NULL;
}

/* IsConstant - check to see if the value of a symbol is a constant */
int IsConstant(Symbol *symbol)
{
    return symbol->storageClass == SC_CONSTANT;
}

/* IsRegister - check to see if the value of a symbol is a register */
int IsRegister(Symbol *symbol)
{
    return symbol->storageClass == SC_REGISTER;
}

/* DumpSymbols - dump a symbol table */
void DumpSymbols(SymbolTable *table, char *tag)
{
    Symbol *sym;
    if ((sym = table->head) != NULL) {
        VM_printf("%s:\n", tag);
        for (; sym != NULL; sym = sym->next)
            VM_printf("  %c %c %08x %08x %s\n", "CVR"[sym->storageClass], "ISAF"[sym->type->id], sym->offset, sym->initialValue, sym->name);
    }
}
