/* db_vmheap.c - heap manager
 *
 * Copyright (c) 2011 by David Michael Betz.  All rights reserved.
 *
 */

#include <string.h>
#include "db_system.h"
#include "db_vm.h"
#include "db_vmdebug.h"

/* round to multiple of the word size */
#define WORDSIZE(n)     (((n) + ALIGN_MASK) & ~ALIGN_MASK)

/* InitHeap - initialize a string heap */
int InitHeap(System *sys, ImageHdr *image, VMUVALUE dataSize, VMUVALUE nHandles)
{
    size_t heapSize = dataSize + nHandles * sizeof(ObjectHandle);
    ObjectHandle *handle;

    /* allocate space for the heap */
    if (!(image->data = AllocateFreeSpace(sys, heapSize)))
        return VMFALSE;
        
    /* setup the handle array */
    image->handles = (ObjectHandle *)(image->data + dataSize);
    image->endHandles = image->handles + nHandles;
    image->nHandles = nHandles;
    
    /* create the handle free list */
    image->freeHandles = NULL;
    for (handle = image->handles; handle < image->endHandles; ++handle) {
        handle->hdr = (ObjectHdr *)image->freeHandles;
        image->freeHandles = handle;
    }

    /* setup the free data space */
    image->free = image->data;
    
    /* return successfully */
    return VMTRUE;
}

/* ReserveHeapSpace - reserve some of the heap space for preallocated objects */
void ReserveHeapSpace(ImageHdr *image, VMUVALUE dataSize, VMUVALUE nHandles)
{
    for (; nHandles > 0; --nHandles)
        image->freeHandles = (ObjectHandle *)image->freeHandles->hdr;
    image->free += dataSize;
}

/* StringAlloc - allocate a new string */
VMVALUE StringAlloc(ImageHdr *image, VMUVALUE length)
{
    VMUVALUE size = sizeof(ObjectHdr) + WORDSIZE(length);
    ObjectHandle *h;
    ObjectHdr *hdr;

    /* find a free handle */
    if (!(h = image->freeHandles))
        return 0;

    /* remove the handle from the free list */
    image->freeHandles = (ObjectHandle *)h->hdr;

    /* make sure there's enough space */
    if ((uint8_t *)image->handles - image->free < size) {
        CompactHeap(image);
        if ((uint8_t *)image->handles - image->free < size) {
            h->hdr = (ObjectHdr *)image->freeHandles;
            image->freeHandles = h;
            return 0;
        }
    }

    /* allocate the next available block */
    hdr = (ObjectHdr *)image->free;
    image->free += size;
    
    /* store the backpointer and length */
    hdr->handle = h;
    hdr->prototype = PROTO_BVECTOR;
    hdr->size = length;
    
    /* store the pointer from the handle to the heap data */
    h->hdr = hdr;

    /* return successfully */
    return image->endHandles - h;
}

/* GetObjectSize - get the size of a heap object */
VMUVALUE GetObjectSize(ObjectHdr *hdr)
{
    int size = sizeof(ObjectHdr);
    switch (hdr->prototype) {
    case PROTO_CODE:
    case PROTO_BVECTOR:
        size += WORDSIZE(hdr->size);
        break;
    case PROTO_VECTOR:
        size += hdr->size * sizeof(VMVALUE);
        break;
    }
    return size;
}

/* CompactHeap - compact the heap */
void CompactHeap(ImageHdr *image)
{
    uint8_t *data, *next, *free;

    /* initialize the heap scan */
    data = next = image->data;
    free = image->free;

    /* compact the heap */
    while (data < free) {
        ObjectHdr *hdr = (ObjectHdr *)data;
        VMUVALUE size = GetObjectSize(hdr);
        if (hdr->handle) {
            if (data != next) {
                hdr->handle->hdr = (ObjectHdr *)(next + sizeof(ObjectHdr));
                memmove(next, data, size);
            }
            next += size;
        }
        data += size;
    }

    /* store the new free pointer */
    image->free = next;
}

/* DumpHeap - dump the heap */
void DumpHeap(ImageHdr *image)
{
    uint8_t *data, *free;

    /* initialize the heap scan */
    data = image->data;
    free = image->free;

    /* dump the heap */
    while (data < free) {
        ObjectHdr *hdr = (ObjectHdr *)data;
        VMUVALUE allocatedSize = GetObjectSize(hdr);
        uint8_t *addr = GetBVectorBase(hdr);
        VM_printf("%0*x h: %0*x, p: %0*x s: %d",
                  sizeof(VMVALUE) * 2, data - image->data,
                  sizeof(VMVALUE) * 2, hdr->handle ? image->endHandles - hdr->handle : 0,
                  sizeof(VMVALUE) * 2, hdr->prototype,
                  hdr->size);
        if (hdr->prototype == PROTO_BVECTOR) {
            VMUVALUE count;
            VM_printf(" '");
            for (count = hdr->size; count > 0; --count)
                VM_putchar(*addr++);
            VM_putchar('\'');
        }
        else if (hdr->prototype == PROTO_CODE) {
            VM_putchar('\n');
            DecodeFunction(0, addr, hdr->size);
        }
        VM_putchar('\n');
        data += allocatedSize;
    }
}
